/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Gareth Johnson
 */

/**
 * Defines {@link baja.Format}.
 * @module baja/obj/Format
 */
define(["bajaScript/sys",
        "bajaScript/baja/obj/Simple",
        "lex!"],
        function (baja, Simple, lexjs) {

  "use strict";

  var subclass = baja.subclass,
      callSuper = baja.callSuper,
      objectify = baja.objectify,
      strictArg = baja.strictArg,
      bajaDef = baja.def,
      bajaHasType = baja.hasType;

  /**
   * `Format` is used to format `Object`s into `String`s using
   * a standardized formatting pattern language.  The format String is normal
   * text with embedded scripts denoted by the percent (%) character.  Use
   * "%%" to escape a real %.
   *
   * A script is one or more calls chained together using the dot (.) operator.
   * Calls are mapped to methods using the order given below.
   *
   * If a script cannot be processed successfully, an error will be returned.
   *
   * To define an alternate output to use if an error is encountered, include a
   * ? followed by another script within the same % pair.  More than one fallback
   * can be defined by delimiting the fallbacks with a ?.
   *
   * Given the call "foo" the order of attempted resolutions is:
   * <ol>
   *   <li>special call (see below)</li>
   *   <li>getFoo(Context)</li>
   *   <li>foo(Context)</li>
   *   <li>get("foo")</li>
   * </ol>
   *
   * The following special functions are available to use in a script:
   * <ol>
   *   <li>lexicon(module:key)</li>
   *   <li>time() returns the current time as an AbsTime</li>
   *   <li>user() returns gets the current user's name</li>
   * </ol>
   *
   * This Constructor shouldn't be invoked directly. Please use the `make()`
   * methods to create an instance of a `Format`.
   *
   * Examples of formats:
   * <pre>
   * "hello world"
   * "my name is %displayName%"
   * "my parent's name is %parent.displayName%"
   * "%parent.value?lexicon(bajaui:dialog.error)%"
   * "%out.value?out.status?lexicon(bajaui:dialog.error)%"
   * </pre>
   *
   * @example <caption>Formats can use getter functions from an object as well
   * as function names</caption>
   * var obj = {
   *   getFoo: function () {
   *     return {
   *       getBar: function () {
   *         return {
   *           value: function () {
   *             return 3.1415;
   *          }
   *        }
   *       }
   *     }
   *   }
   * }
   *
   * var fmt = baja.Format.make("%foo.bar.value%");
   * return fmt.format( { object: obj } )
   *   .then(function (value) {
   *     // prints 3.1415
   *     console.log(value);
   *   });
   *
   * @class
   * @alias baja.Format
   * @extends baja.Simple
   */
  var Format = function Format(pattern) {
    callSuper(Format, this, arguments);
    this.$pattern = strictArg(pattern, String);
  };

  subclass(Format, Simple);

  /**
   * Default Format instance.
   * @type {baja.Format}
   */
  Format.DEFAULT = new Format("");

  /**
   * Make a `Format`.
   *
   * @param {String} [pattern] the `Format` Pattern `String`.
   * @returns {baja.Format}
   */
  Format.make = function (pattern) {
    pattern = pattern || "";

    if (pattern === "") {
      return Format.DEFAULT;
    }

    strictArg(pattern, String);

    return new Format(pattern);
  };

  /**
   * Make a `Format`.
   *
   * @param {String} [pattern] the `Format` Pattern `String`.
   * @returns {baja.Format}
   */
  Format.prototype.make = function (pattern) {
    return Format.make.apply(Format, arguments);
  };

  /**
   * Decode a `String` to a `Format`.
   *
   * @param {String} str
   * @returns {baja.Format}
   */
  Format.prototype.decodeFromString = function (str) {
    return Format.make(str);
  };

  /**
   * Encode `Format` to a `String`.
   *
   * @returns {String}
   */
  Format.prototype.encodeToString = function () {
    return this.$pattern;
  };

  /**
   * Return a `String` representation of the `Format`.
   *
   * @returns {String}
   */
  Format.prototype.toString = function () {
    return this.$pattern;
  };

  /**
   * Return the inner value of the `Format`.
   *
   * @returns {String}
   */
  Format.prototype.valueOf = Format.prototype.toString;

  // Scope some of the variables here...
  (function () {
    var formatMatches = [],
        formatErr = "error: ",
        lexiconRegex = /^lexicon\(([a-zA-Z0-9]+)\:([a-zA-Z0-9.\-_]+)((?:\:[a-zA-Z0-9\$(?:\s)*]+)*)\)$/;

    // Replace %% with %
    formatMatches.push({
      isMatch: function (content) {
        return content === "";
      },
      replace: function (obj) {
        obj.next("%");
      }
    });

    // Identity %.%
    formatMatches.push({
      isMatch: function (content) {
        return content === ".";
      },
      replace: function (obj) {
        var val,
            parent,
            slot = null,
            i,
            o,
            next = obj.next;

        if (obj.target) {
          if (obj.target.container) {
            if (obj.target.propertyPath && obj.target.propertyPath.length > 0) {
              val = parent = obj.target.container;

              for (i = 0; i < obj.target.propertyPath.length; ++i) {
                slot = obj.target.propertyPath[i];
                parent = val;
                val = val.get(slot);
              }

              next(obj.display ? parent.getDisplay(slot) : parent.get(slot));
            }
            else {
              next(obj.display ? obj.target.container.getDisplay(obj.target.slot) : obj.target.container.get(obj.target.slot));
            }
          }
          else {
            o = obj.target.getObject();
            next(obj.display && bajaHasType(o) && o.getType().isComplex() ? o.getDisplay() : o);
          }
        }
        else {
          o = obj.object;
          next(obj.display && bajaHasType(o) && o.getType().isComplex() ? o.getDisplay() : o);
        }
      }
    });

    formatMatches.push({
      isMatch: function (content) {
        return content === "user()";
      },
      replace: function (obj) {
        obj.next(baja.getUserName());
      }
    });

    // Lexicon %.%
    formatMatches.push({
      isMatch: function (content) {
        return content.match(lexiconRegex);
      },
      replace: function (obj) {
        var result = lexiconRegex.exec(obj.content);
        lexiconRegex.lastIndex = 0;

        // Asynchronously request the lexicon value
        lexjs.module(result[1])
          .then(function (lex) {
            if (result.length > 3 && result[3]) {
              var fmtArgs = result[3].substring(1).split(":")
                .map(function(fmtArg){
                  return baja.SlotPath.unescape(fmtArg);
                });
              obj.next(lex.get({"key": result[2], "args": fmtArgs}));
            }
            else {
              obj.next(lex.get(result[2]));
            }
          }, function (err) {
            obj.next(null, formatErr + " : " + err + " + " + obj.content);
          });
      }
    });

    // Reflect Call
    formatMatches.push({
      isMatch: function (content) {
        // Always last in the list so this should always match
        return true;
      },
      replace: function (obj) {
        var val,
            parent,
            slot = null,
            x,
            next = obj.next,
            split = obj.content.split(/\./g);

        // Process the format text
        val = parent = obj.object;

        // TODO Eventually, instead of iterating over split,
        // need to use reduce, with val being resolved with each iteration.
        // This will allow asynchronous evaluations (such as with
        // time().toDateString)

        for (x = 0; x < split.length; ++x) {

          if (split[x] === "time()") {
            val = baja.AbsTime.now();
            continue;
          }

          if (val === null || val === undefined) {
            next(null, formatErr + obj.content);
            return;
          }

          // First try looking for the Slot
          if (bajaHasType(val, "baja:Complex") && val.has(split[x])) {
            slot = val.getSlot(split[x]);
            parent = val;
            val = val.get(slot);
          }
          // If there's not Slot then see if a function exists
          else {
            // Nullify this since at this point we're no longer looking up a Slot chain
            slot = null;
            parent = null;
            //If pattern starts with . returns the object itself for first empty split
            if(split[x] === ""){
              val = obj.object;
            }
            else if (typeof val["get" + split[x].capitalizeFirstLetter()] === "function") {
              val = val["get" + split[x].capitalizeFirstLetter()]();
            }
            else if (typeof val[split[x]] === "function") {
              val = val[split[x]]();
            }
            else if (typeof val.get === "function") {
              val = baja.def(val.get(split[x]), null);
            }
            else {
              val = null;
            }
          }
        }

        if (slot && parent) {
          val = obj.display ? parent.getDisplay(slot) : parent.get(slot);
          next(val);
          return;
        }

        if (bajaHasType(val, "baja:Complex")) {
          parent = val.getParent();
          if (parent) {
            val = obj.display ? parent.getDisplay(val.getPropertyInParent()) : parent.get(val.getPropertyInParent());
            next(val);
            return;
          }
        }
        // As a last resort, just call toString
        if (val !== null) {
          next(val.toString());
          return;
        }
        next(val);
        return;
      }
    });


    /**
     * Format the specified object using the format pattern.
     *
     * This method can take an object literal or a single pattern `String`
     * argument.
     *
     * @param {Object} obj
     * @param {String} obj.pattern the format pattern to process.
     * @param {Object|baja.Component} [obj.object] JavaScript Object or baja:Component
     * referenced by the format scripts.
     * @param {Boolean} [obj.display] if true, the display string of a Property value is used.
     *                                If false, the `toString` version of a Property value is used.
     *                                By default, this value is true (in BajaScript, most of the time
     *                                we're dealing with mounted Components in a Proxy Component Space).
     * @param {Function} [obj.ok] (Deprecated: use Promise) the ok callback
     * called when the Format string has been processed. The resultant String
     * will be passed to this function as an argument
     * @param {Function} [obj.fail] (Deprecated: use Promise) the fail callback
     * called if there's a fatal error processing the format.
     * @returns {Promise.<string>}
     */
    Format.format = function (obj) {

      // TODO: Currently format processing doesn't work in exactly the same way as Niagara.
      // Certainly it can never be 100% accurate. However, we can try to cover most common use cases
      // that we want to support.

      obj = objectify(obj, "pattern");

      var pattern = obj.pattern,
          regex = /%[^%]*%/g,
          matches = [],
          res,
          index = -1,
          content,
          splitContent = [],
          splitIndex = 0,
          process,
          tryNext,
          cb = new baja.comm.Callback(obj.ok, obj.fail || baja.fail);

      obj.display = bajaDef(obj.display, true);

      // Find all the matches in the string to process
      for(;;) {
        res = regex.exec(pattern);
        if (res) {
          // Add data (remove start and end % characters)
          matches.push(res[0].substring(1, res[0].length - 1));
        }
        else {
          break;
        }
      }

      /**
       * Called to process the next pattern match
       *
       * @param {string} res results of the previous match, null if an error occurred
       * @param {string} [errMsg] error to use if an error occurred
       */
      function next(res, errMsg) {
        var isError = res === null;
        matches[index] = isError ? (errMsg ? errMsg : formatErr + obj.content) : res;

        try {
          if (isError && splitIndex + 1 < splitContent.length) {
            tryNext();
          } else {
            process();
          }
        }
        catch(err) {
          cb.fail(err);
        }
      }
      obj.next = next;

      // Called when the format has finished processing
      function finish() {
        // Reset the regex expression object
        regex.lastIndex = 0;

        // Form the new string and pass the result
        var i = 0;
        cb.ok(pattern.replace(regex, function () {
          return matches[i++];
        }));
      }

      // Process the next statement (this will get called after each pattern is replaced
      process = function () {
        if (++index > (matches.length - 1)) {
          finish();
          return;
        }

        content = obj.content = matches[ index ];
        splitContent = obj.splitContent = content.split('?');
        splitIndex = -1;

        tryNext();
      };

      tryNext = function () {
        var nextContent = obj.content = String(splitContent[++splitIndex]);

        // Match up with a format handler and process it
        var i;
        for (i = 0; i < formatMatches.length; ++i) {
          // If we have a match then process it
          if (formatMatches[i].isMatch(nextContent)) {

            formatMatches[i].replace(obj);
            return;
          }
        }

        // If we get here then we couldn't find a match
        next(null);
      };

      try {
        process();
      }
      catch(err) {
        cb.fail(err);
      }

      return cb.promise();
    };
  }());

  /**
   * Format the specified object using the format pattern.
   *
   * @see baja.Format.format
   * 
   * @param {Object} [obj]
   * @param {Function} [obj.ok] (Deprecated: use Promise) the ok callback called
   * when the Format string has been processed. The resultant String will be
   * passed to this function as an argument
   * @param {Function} [obj.fail] (Deprecated: use Promise) the fail callback
   * called if there's a fatal error processing the format.
   * @returns {Promise.<String>}
   */
  Format.prototype.format = function (obj) {
    obj = objectify(obj);
    obj.pattern = this.$pattern;
    return Format.format(obj);
  };

  return Format;
});