/**
 * @copyright 2015 Tridium, Inc. All Rights Reserved.
 * @author Gareth Johnson
 */

/**
 * Defines `OrdTarget` (not exposed on `baja` namespace).
 * @module baja/ord/OrdTarget
 */
define(["bajaScript/sys"], 
    function (baja) {
  
  "use strict";
  
  var subclass = baja.subclass,
      BaseBajaObj = baja.BaseBajaObj;
  
  /**
   * ORD Target.
   * 
   * This constructor shouldn't be invoked directly.
   *  
   * @class
   * @alias module:baja/ord/OrdTarget
   * @extends baja.BaseBajaObj
   * @inner
   * @public
   *
   * @param {module:baja/ord/OrdTarget} [base]  the base ORD Target
   */
  var OrdTarget = function OrdTarget(base) {    
    this.base = base || null;  
    this.object = null;
    
    if (base && typeof base === "object") {
      base.next = this;
    }    
  };
  
  subclass(OrdTarget, BaseBajaObj);
  
  /**
   * Return the Component for the ORD Target.
   *
   * @returns {baja.Component}
   */
  OrdTarget.prototype.getComponent = function () {
    var that = this,
        container = that.container,
        object = that.object,
        base = that.base,
        o;
    
    if (baja.hasType(container) && container.getType().isComponent()) {
      return container;
    }
    
    if (baja.hasType(object)) {
      if (object.getType().isComponent()) {
        return object;
      }
      
      if (object.getType().isComplex()) {
        o = object.getParent();
        while (o !== null) {
          if (o.getType().isComponent()) {
            break;
          }
          o = o.getParent();
        }
        if (o !== null) {
          return o;
        }
      }
    }
    
    if (base && typeof base === "object") {
      return base.getComponent();
    }
    
    return null;
  };  
  
  /**
   * Return the object associated with this `OrdTarget`.
   * 
   * This method will attempt to access any resolved value from its
   * Property. If the resolved value wasn't resolved from a Property then 
   * the original resolved object is returned.
   *
   * @returns {baja.Object}
   */
  OrdTarget.prototype.getObject = function () {
    var that = this,
        parent = that.container,
        slot = that.slot,
        propertyPath = that.propertyPath,
        i,
        val;
      
    if (parent) {
      // Walk up the Property Path
      if (propertyPath) {
        if (propertyPath.length > 0) {
          val = parent;
          for (i = 0; i < propertyPath.length; ++i) {
            val = val.get(propertyPath[i]);
          } 
          return val;          
        }
      }
      // If no PropertyPath then access from the Slot
      else if (slot && slot.isProperty()) {
        return parent.get(slot);
      }
    }
    
    // By default, just return the object originally resolved
    return that.object;
  }; 
  
  return OrdTarget;
});