/**
 * @copyright 2017 Tridium, Inc. All Rights Reserved.
 * @author Gareth Johnson
 */

/**
 * Core communication base class.
 *
 * @module env/Connection
 * @private
 */
define([ "bajaScript/comm", "bajaPromises" ], function (baja, Promise) {

  "use strict";

  /**
   * The abstract base class for all full-duplex browser connections.
   *
   * @class
   * @private
   * @alias module:env/Connection
   * @see env/ConnectionManager
   * @param {Object} handlers Holds a collection of functions used as asynchronous callbacks.
   */
  var Connection = function (handlers) {
    var that = this;

    that.$internalConnection = null;  
    that.$handlers = handlers;
  };

  var WS_OPEN = Connection.WS_OPEN = 1;

  /**
   * Authenticate with the server so that the connection can be started.
   * 
   * @private
   * @returns {Promise|*}
   */
  Connection.prototype.authenticate = function () {
    var that = this;
    return Promise.resolve(that.doAuthenticate())
      .then(function (res) {
        that.$authenticated = true;
        return res;
      });
  };

  /**
   * Called to authenticate with the server prior to starting the connection.
   * By default, does nothing. Override this function; call authenticate().
   *
   * @private
   * @abstract
   * @returns {Promise|*}
   */
  Connection.prototype.doAuthenticate = function () {
    
  };

  /**
   * Called to start the connection. By default, does nothing.
   * 
   * @private
   * @abstract
   * @returns {Promise|*}
   */
  Connection.prototype.start = function () {
    
  };

  /**
   * Called to close the connection.
   *
   * @private
   */
  Connection.prototype.close = function () {
    var that = this;

    if (that.$internalConnection && 
      typeof that.$internalConnection.close === "function") {
      that.$internalConnection.close();
    }
  };

  /**
   * Called to send some data to the connection.
   *
   * @private
   * 
   * @param {Object} frameData The frame data to send.
   */
  Connection.prototype.send = function (frameData) {
    var that = this;

    if (!that.isOpen()) {
      throw new Error("Connection unavailable!");
    }

    // Convert the frame to a string and send it down the connection.
    that.$internalConnection.send(frameData.frame.toString());
  };

  /**
   * Returns the internal raw underlying connection object or null
   * if there isn't one.
   *
   * @private
   * 
   * @returns The internal connection or null.
   */
  Connection.prototype.getInternalConnection = function () {
    return this.$internalConnection;
  };

  /**
   * Return true if the underlying connection is currently in an open state.
   *
   * @private
   * 
   * @return {Boolean} Returns true if open.
   */
  Connection.prototype.isOpen = function () {
    return !!(this.$authenticated &&
      this.$internalConnection &&
      this.$internalConnection.readyState === WS_OPEN);
  };

  return Connection;
});
